/*
 *  $Id: file.c,v 1.42 2000/02/27 18:12:15 schaefer Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#include <fist.h>
#include <mfs.h>
#include "mfs_definitions.h"
#include "mfs_interface.h"
#include "mfs_com_interface.h"

/*******************
 * File Operations *
 *******************/

// STATIC int
// mfs_lseek(inode_t *inode, file_t *file, off_t offset, int origin)
STATIC loff_t
mfs_llseek(file_t *file, loff_t offset, int origin)
{
    loff_t err;
    file_t *hidden_file = ftohf(file);

    print_entry_location();

    fist_dprint(6, "mfs_llseek: file=%x, offset=0x%x, origin=%d\n",
		(int) file, (int) offset, origin);

    //    fist_print_file("LLSEEK Before file", file);
    //    fist_print_file("LLSEEK Before hidden_file", hidden_file);

    // ION
    /* always set hidden position to this one */
    hidden_file->f_pos = file->f_pos;
    if (file->f_reada) { /* update readahead information if needed */
	hidden_file->f_reada = file->f_reada;
	hidden_file->f_ramax = file->f_ramax;
	hidden_file->f_raend = file->f_raend;
	hidden_file->f_ralen = file->f_ralen;
	hidden_file->f_rawin = file->f_rawin;
    }
    ASSERT(hidden_file->f_reada == file->f_reada);
    ASSERT(hidden_file->f_ramax == file->f_ramax);
    ASSERT(hidden_file->f_raend == file->f_raend);
    ASSERT(hidden_file->f_ralen == file->f_ralen);
    ASSERT(hidden_file->f_rawin == file->f_rawin);

    if (hidden_file->f_op && hidden_file->f_op->llseek) {
	err = hidden_file->f_op->llseek(hidden_file, offset, origin);
    } else {
	err = default_llseek(hidden_file, offset, origin);
    }
    if (err < 0) {
	goto out;
    }

    if (err != file->f_pos) {
	file->f_pos = err;
	// ION maybe this?
	// 	file->f_pos = hidden_file->f_pos;
	file->f_reada = 0;
	file->f_version = ++event;
    }

out:
    //    fist_print_file("LLSEEK After file", file);
    //    fist_print_file("LLSEEK After hidden_file", hidden_file);
    print_exit_status(err);
    return err;
}


// STATIC int
// mfs_read(inode_t *inode, file_t *file, char *buf, int count)
STATIC ssize_t
mfs_read(file_t *file, char *buf, size_t count, loff_t *ppos)
{
    int err = -EINVAL;
    file_t *hidden_file = ftohf(file);
    char *hidden_buf;
    mm_segment_t old_fs;
    loff_t pos = *ppos;

    print_entry_location();

    if (!hidden_file->f_op || !hidden_file->f_op->read)
	goto out;
    hidden_buf = vmalloc(count);

    old_fs = get_fs();
    set_fs(KERNEL_DS);

    err = hidden_file->f_op->read(hidden_file, hidden_buf, count, &pos);

    set_fs(old_fs);
    if (err > 0) {
	mfs_decode_block(hidden_buf, hidden_buf, err);
	fist_copy_attr_atime(file->f_dentry->d_inode,
			     hidden_file->f_dentry->d_inode);
	if (copy_to_user(buf, hidden_buf, err))
	    err = -EINVAL;
    }

    // MAJOR HACK
    /*
     * because pread() does not have any way to tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on read() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;

    vfree(hidden_buf);

out:
    print_exit_status(err);
    return err;
}


// STATIC int
// mfs_write(inode_t *inode, file_t *file, const char *buf, int count)
#if 0
STATIC ssize_t
mfs_write(file_t *file, const char *buf, size_t count, loff_t *ppos)
{
    int err = -EINVAL;
    file_t *hidden_file = ftohf(file);
    char *hidden_buf;
    mm_segment_t old_fs;
    loff_t pos = *ppos;

    if (!hidden_file->f_op || !hidden_file->f_op->write)
	goto out;

    if (file->f_flags & O_APPEND)
	pos = file->f_dentry->d_inode->i_size;

    hidden_buf = vmalloc(count);
    copy_from_user(hidden_buf, buf, count);
    mfs_encode_block(hidden_buf, hidden_buf, count);

    old_fs = get_fs();
    set_fs(KERNEL_DS);

    err = hidden_file->f_op->write(hidden_file, hidden_buf, count, &pos);

    set_fs(old_fs);

    // MAJOR HACK
    /*
     * because pread() does not have any way to tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on read() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;

    /* writing may increase the file's size */
    if (err > 0)
	fist_copy_attr_timesizes(file->f_dentry->d_inode,
				 hidden_file->f_dentry->d_inode);

    vfree(hidden_buf);

out:
    print_exit_status(err);
    return err;
}
#else
/*
 * Adapted from 2.1 generic_file_write
 */
// STATIC int
// mfs_write(inode_t *inode, file_t *file, const char *buf, int count)
STATIC ssize_t
mfs_write(file_t *file, const char *buf, size_t count, loff_t *ppos)
{
    file_t *hidden_file;
    page_t *page, **hash;
    unsigned long page_cache = 0;
    unsigned long pgpos, orig_pgpos, offset;
    unsigned long bytes, written;
    size_t bytes_to_write;
    loff_t pos, orig_pos, tmp_pos;
    ssize_t status;
    long didread;
    mm_segment_t old_fs;
    char *hidden_buffer;
    inode_t *inode = file->f_dentry->d_inode;
    inode_t *hidden_inode = itohi(inode);

    /* @@@ */
    ASSERT(file != NULL);
    hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);
	return hidden_file->f_op->write(hidden_file,
					  buf,
					  count,
					  ppos);
    /* @@@ */

    print_entry_location();

    /* check that offsets are correct */
    ASSERT(file != NULL);
    ASSERT(ppos != NULL);

    print_entry_location();

    pos = *ppos;
redo:
    written = 0;
    status = 0;

    old_fs = get_fs();
    hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    hidden_buffer = kmalloc(PAGE_SIZE, GFP_KERNEL);
    if (!hidden_buffer) {
	printk("Out of memory.\n");
	status = -ENOMEM;
	goto out;
    }
    /* we don't want anybody to do updates while we write, so lock the inode */
    down(&hidden_inode->i_sem);

    if (file->f_flags & O_APPEND)
	pos = inode->i_size;

    fist_dprint(6, "EZK: WRITE: initial values: pos %ld, count %d\n", (long) pos, count);
    fist_print_file("EZK: WRITE file IN", file);

    /* we need to handle sparse files differently */
    orig_pos = pos;
    orig_pgpos = orig_pos & PAGE_MASK;
    if (pos > inode->i_size)
	pos = inode->i_size;

    while (count) {
	fist_dprint(7, "WRITE: pos %ld, written %ld, count %d, buf 0x%lx\n",
		    (long) pos, written, count, (long) buf);
	/*
	 * Try to find the page in the cache. If it isn't there,
	 * allocate a free page.
	 */
	offset = (pos & ~PAGE_MASK);
	pgpos = pos & PAGE_MASK;

	if ((bytes = PAGE_SIZE - offset) > count && pgpos >= orig_pgpos)
	    bytes = count;

	hash = page_hash(inode, pgpos);
	if (!(page = __find_page(inode, pgpos, *hash))) {
	    if (!page_cache) {
		page_cache = __get_free_page(GFP_KERNEL);
		if (page_cache)
		    continue;
		status = -ENOMEM;
		break;
	    }
	    page = mem_map + MAP_NR(page_cache);
	    add_to_page_cache(page, inode, pgpos, hash);
	    page_cache = 0;
	}
	/*
	 * Note: setting of the PG_locked bit is handled
	 * below the i_op->xxx interface.
	 */
	didread = 0;
    page_wait:
	wait_on_page(page);
	if (PageUptodate(page))
	    goto do_update_page;

	/*
	 * The page is not up-to-date ... if we're writing less
	 * than a full page of data, we may have to read it first.
	 * But if the page is past the current end of file, we must
	 * clear it before updating.
	 */
	if (bytes < PAGE_SIZE) {
	    if (pgpos < inode->i_size) {
		status = -EIO;
		if (didread >= 2)
		    goto done_with_page;
		/* call our own readpage function */
		status = inode->i_op->readpage(file, page);
		if (status < 0)
		    goto done_with_page;
		didread++;
		goto page_wait;
	    } else {
		/* Must clear for partial writes */
		fist_dprint(7, "WRITE1: clearing page at offset 0x%x\n", (int) pgpos);
		memset((void *) page_address(page), 0, PAGE_SIZE);
	    }
	} else if (pos < orig_pos) {
	    /* Must clear for sparse files */
	    fist_dprint(7, "WRITE2: clearing page at offset 0x%x\n", (int) pgpos);
	    memset((void *) page_address(page), 0, PAGE_SIZE);
	}
	/*
	 * N.B. We should defer setting PG_uptodate at least until
	 * the data is copied. A failure in i_op->updatepage() could
	 * leave the page with garbage data.
	 */
	set_bit(PG_uptodate, &page->flags);

    do_update_page:
	/* Alright, the page is there.  Now update it. */
	fist_dprint(7, "WRITE: count %d, pos %ld, offset %ld, pgpos 0x%x, bytes %d\n",
		    count, (int) pos, (int) offset, (int) pgpos, (int) bytes);

	if (pgpos < orig_pgpos)
	    /* nothing to do, not even copy_from_user */
	    goto encode_page;
	if (pgpos == orig_pgpos) {
	    /*
	     * this is the only interesting case,
	     * we have to shift from orig_pos to pos somehow
	     */
	    pos = orig_pos;
	    offset = pos & ~PAGE_MASK;
	    if ((bytes = PAGE_SIZE - offset) > count)
		bytes = count;
	}
#if 0
	status = inode->i_op->updatepage(file, page, buf,
					 offset, bytes, sync);
#else
	fist_dprint(7, "WRITE: copying %ld bytes at offset %ld\n", bytes, offset);
	copy_from_user((char *) (page_address(page) + offset), buf, bytes);
    encode_page:
	mfs_encode_block((char *) page_address(page), hidden_buffer, PAGE_SIZE);

	hidden_file->f_pos = pgpos; /* is this needed? */
	if (inode->i_size < pgpos + PAGE_SIZE)
	    if (inode->i_size > pos + bytes)
		bytes_to_write = inode->i_size - pgpos;
	    else
		bytes_to_write = offset + bytes;
	else
	    bytes_to_write = PAGE_SIZE;

	fist_dprint(7, "WRITE: writing at pgpos %ld, size %ld\n", pgpos, bytes_to_write);
	/* switch to kernel space */
	set_fs(KERNEL_DS);
	tmp_pos = pgpos;
	status = hidden_file->f_op->write(hidden_file,
					  hidden_buffer,
					  bytes_to_write,
					  &tmp_pos);
	fist_dprint(7, "WROTE: tmp_pos %ld, status %d\n", (int) tmp_pos, (int) status);

	/* switch back to user space */
	set_fs(old_fs);
#endif

	/* the rest of the code must not see that we are writing extra bytes */
	/* do not adjust status if only filling up holes */
	if (status > 0 && pgpos >= orig_pgpos) {
	    if (status < offset) {
		status = 0;
            }
	    else {
		status -= offset;
		if (status > bytes)
		    status = bytes;
	    }
        }
    done_with_page:
	__free_page(page);
	if (status < 0)
	    break;

	/* do not adjust these variables if only filling up holes */
	if (pgpos >= orig_pgpos) {
	    written += status;
	    count -= status;
	    buf += status;
	}
	pos += status;
    }
    // MAJOR HACK
    /*
     * because pread() does not have any way to tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on read() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;
    //    file->f_pos = pos;

    if (pos > inode->i_size)
	inode->i_size = pos;

    up(&hidden_inode->i_sem);

    kfree_s(hidden_buffer, PAGE_SIZE);

    if (page_cache)
	free_page(page_cache);

    status = written ? written : status;

    if (written < count) {
       mfs_status_t mfs_status;

       printk("mfs_write: wanted to write %d, only wrote %ld [0x%x].\n",
              count, written, status);

       if (status == -ENOSPC) {
          printk("mfs_write: -ENOSPC; SYSTEM_LOW_IN_SPACE.\n");

	  /* TODO @@@
	   * In case of failures of the underlying fs due to -ENOSPC,
	   * we can initiate a migration and be awaken again and retry
	   * afterwards. If file was open with O_NBLOCK/NODELAY, in that case
	   * we fail.
	   */

	  mfs_status = mfs_do_request(stopd(inode->i_sb)->mfs_com_private,
				      MFS_COMMAND_SYSTEM_LOW_IN_SPACE,
				      NULL,
				      NULL,
				      NULL);
	  if (mfs_status == MFS_STATUS_OK) {
	     printk("mfs_write: continuing write where we left off.\n");
             count = count - written;
             buf += written;
             pos += written;
             /* @@@ not sure this is right, especially with ppos @@@ */
	     goto redo;
	  }
	  else {
	    printk("mfs_write: uh? mfsd answering a migrate with 0x%x.\n",
		   status);
	  }
       }
    }

out:
    print_exit_status(status);
    //    fist_print_file("WRITE file", file);
    //    fist_print_file("WRITE hidden_file", hidden_file);
    //    fist_set_debug_value(0);
    return status;
}
#endif

// STATIC int
// mfs_readdir(inode_t *inode, file_t *file, void *dirent, filldir_t filldir)
STATIC int
mfs_readdir(file_t *file, void *dirent, filldir_t filldir)
{
    int err = -ENOTDIR;
    file_t *hidden_file = ftohf(file);
    inode_t *inode = file->f_dentry->d_inode;
    inode_t *hidden_inode = itohi(inode);

    print_entry_location();

    fist_checkinode(inode, "mfs_readdir");
    if (!hidden_file || !hidden_file->f_op || !hidden_file->f_op->readdir)
	goto out;

    down(&hidden_inode->i_sem);
    err = hidden_file->f_op->readdir(hidden_file, dirent, filldir);
    up(&hidden_inode->i_sem);

    file->f_pos = hidden_file->f_pos;
    if (err > 0)
	fist_copy_attr_atime(file->f_dentry->d_inode,
			     hidden_file->f_dentry->d_inode);
    fist_checkinode(inode, "post mfs_readdir");

out:
    print_exit_status(err);
    return err;
}


// STATIC int
// mfs_select(inode_t *inode, file_t *file, int flag, select_table * wait)
STATIC unsigned int
mfs_poll(file_t *file, poll_table *wait)
{
    int err = 0;
    file_t *hidden_file = ftohf(file);

    print_entry_location();

    /* NOTES
     *    - if this is the special file, and the current process is
     *      mfsd, this should be handled specially.
     */
    if (!mfs_is_special_poll(file,
                             wait,
                             stopd(file->f_dentry->d_inode->i_sb)
                                ->mfs_com_private,
                             &err /* mask, infact */)) {
       if (!hidden_file->f_op || !hidden_file->f_op->poll)
	   goto out;

       err = hidden_file->f_op->poll(hidden_file, wait);
   }

out:
    print_exit_status(err);
    return err;
}


STATIC int
mfs_ioctl(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
    int err = 0, val;
    file_t *hidden_file;

    print_entry_location();

    /* check if asked for local commands */
    switch (cmd) {
#ifdef FIST_DEBUG
    case FIST_IOCTL_GET_DEBUG_VALUE:
	val = fist_get_debug_value();
	fist_dprint(6, "IOCTL GET: got arg %d\n", val);
	err = put_user(val, (int *) arg);
	break;

    case FIST_IOCTL_SET_DEBUG_VALUE:
	err = get_user(val, (int *) arg);
	if (err)
	    break;
	fist_dprint(6, "IOCTL SET: got arg %d\n", val);
	if (val < 0 || val > 20) {
	    err = -EINVAL;
	    break;
	}
	fist_set_debug_value(val);
	break;
#endif /* FIST_DEBUG */

    case MFS_IOCTL_DEBUG:
       if (!suser()) {
          err = -EPERM;
       }
       else if (arg == 1) {
          printk("ufs\n");
          mfs_dump_all_fs_files(itohi(inode)->i_sb);
          err = 0;
       }
       else if (arg == 2) {
          printk("mfs\n");
          mfs_dump_all_fs_files(inode->i_sb);
          err = 0;
       }
       else if (arg == 3) {
          printk("ufs invalidate_inodes()\n");
          invalidate_inodes(itohi(inode)->i_sb);
          err = 0;
       }
       else if (arg == 4) {
          printk("mfs invalidate_inodes()\n");
          invalidate_inodes(inode->i_sb);
          err = 0;
       }
       else {
          err = -EINVAL;
       }
       break;

    /* BUGS
     *    - We assume only one mfsd process, and that information is stable
     *      Especially for SMP systems we may have to protect with a lock.
     *    - Could use new capabilities or existing ones.
     */

    case MFS_IOCTL_NOTIFY_LIVES: /* mfsd */
       fist_dprint(6, "mfs: MFS_IOCTL_NOTIFY_LIVES.\n");
       err = mfs_notify_lives(file->f_dentry,
                              stopd(inode->i_sb)->mfs_com_private,
                              (void *) arg);
       break;

    /* The following mfsd function verify internally that
     * MFS_IOCTL_NOTIFY_LIVES was called. When the file is closed,
     * the link will be broken and all requests cancelled.
     */

    case MFS_IOCTL_GET_REQUEST_BLOCK: /* mfsd */
       fist_dprint(6, "mfs: MFS_IOCTL_GET_REQUEST_BLOCK.\n");

       if (!suser()) {
          err = -EPERM;
       }
       else {
	  err = mfs_transmit_request(stopd(inode->i_sb)->mfs_com_private,
				     (void *) arg,
                                     file->f_dentry); /* sleeps */
       }
       break;
    case MFS_IOCTL_SEND_ANSWER_BLOCK: /* mfsd */
       fist_dprint(6, "mfs: MFS_IOCTL_SEND_ANSWER_BLOCK.\n");

       if (!suser()) {
          err = -EPERM;
       }
       else {
          err = mfs_receive_answer(stopd(inode->i_sb)->mfs_com_private,
                                   (void *) arg,
                                   file->f_dentry); /* sleeps */
       }
       break;
    case MFS_IOCTL_ATOMIC_MIGRATE: /* mfsd / mfs_migrd */
       fist_dprint(6, "mfs: MFS_IOCTL_ATOMIC_MIGRATE.\n");

       if (!suser()) {
          err = -EPERM;
       }
       else {
          mfs_atomic_migrate_t *buffer
             = (mfs_atomic_migrate_t *)
                  kmalloc(sizeof(mfs_atomic_migrate_t), GFP_KERNEL);

          /* NOTES
           *    - The only constraint is that the check for bit unset, then
           *      check for opened file then set the attribute bit be atomic.
           *    - This allows for one or more migration processes on the
           *      same files.
           * BUGS
           *    - If the migration process is killed in between, there could
           *      be an incoherency. To fix this, we should on mfsd re-start
           *      check for any migrated file of size > 0 not in the database,
           *      and think a lot on the way to make it 100% safe.
           */

          if (buffer) {
             if (copy_from_user((void *) buffer,
                                (void *) arg,
                                sizeof(mfs_atomic_migrate_t))) {
                err = -EFAULT;
             }
             else {
                dentry_t *temp_dentry;
                buffer->path_name[MAX_PATHNAME_LENGTH - 1] = '\0';

		/* NOTES
		 *    - It could be easier if mfsd had the file open, and we
		 *      would check for count == 1. However, that would defeat
		 *      things like O_EXCL. Think again.
		 *    - It is guaranteed that this ioctl() will not
		 *      cause a mfs request (else we would deadlock ...).
		 */

                /* NOTES
                 *    - This is taken from inode.c, and is supposedly right.
                 */
                dget(inode->i_sb->s_root);
                temp_dentry = lookup_dentry(buffer->path_name,
                                            inode->i_sb->s_root,
                                            0);
                if (IS_ERR(temp_dentry)) {
                   printk("mfs: lookup_dentry() error.\n");
                   err = PTR_ERR(temp_dentry);
                }
		else if (temp_dentry->d_inode) {
      		   dentry_t *hidden_dentry = mfs_hidden_dentry(temp_dentry);

                   if (hidden_dentry == NULL) {
		      printk("mfs: MFS_IOCTL_ATOMIC_MIGRATE no underlying.\n");
                      err = -EBUSY; /* no hidden_dentry */
                   }
		   else if (MFS_IS_MIGRATED_FLAG(hidden_dentry->d_inode)) {
		      printk("mfs: MFS_IOCTL_ATOMIC_MIGRATE migrated file.\n");
		      err = -EACCES; /* file is migrated already */
		   }
		   else if (IS_RDONLY(temp_dentry->d_inode)) {
                      printk("mfs: read only fs.\n");
		      err = -EROFS; /* read-only fs */
		   }
                   else if (!S_ISREG(temp_dentry->d_inode->i_mode)) {
                      printk("mfs: not a plain file.\n");
		      err = -EPIPE;
                   }
		   else {
                      /* Is the file open (atomic test) ? We only
                       * check for opens through the mfs filesystem,
                       * not directly through the underlying
                       * fs.
                       */
                      if (mfs_dentry_in_file(temp_dentry)) {
                         printk("mfs: file is open.\n");
                         err = -EPERM; /* file is open */
                      }
                      else {
                         MFS_SET_MIGRATED_FLAG(temp_dentry);
                         MFS_SET_MIGRATED_FLAG(hidden_dentry);
                      }
                   }
                   dput(temp_dentry);
		}
		else {
		   err = -EIO; /* file no longer exists (under that name) */
                   printk("mfs: unexistant file %s.\n", buffer->path_name);
                   dput(temp_dentry);
		}
             }
             kfree(buffer);
          }
          else {
             err = -ENOMEM;
          }
       }
       break;

    case MFS_IOCTL_IMPORT_OBJECT: /* mfs_import */
       fist_dprint(6, "mfs: MFS_IOCTL_IMPORT_OBJECT.\n");
       if (!suser()) {
          err = -EPERM;
       }
       else {
          mfs_status_t status;
          mfs_import_t *mfs_import;

          mfs_import
             = (mfs_import_t *) kmalloc(sizeof(mfs_import_t), GFP_KERNEL);
          if (mfs_import) {
	     if (copy_from_user((void *) mfs_import,
			        (void *) arg,
			        sizeof(mfs_import_t))) {
		err = -EFAULT;
	     }
             else {
		status = mfs_do_request(stopd(file->f_dentry->d_inode->i_sb)
					->mfs_com_private,
					MFS_COMMAND_CREATE_MIGRATED_OBJECT,
					NULL,
					NULL,
					mfs_import);
		if (status == MFS_STATUS_OBJECT_EXISTS) {
		   err = -EEXIST;
		}
		else if (status != MFS_STATUS_OK) {
		   printk("mfs_ioctl: MFS_IOCTL_IMPORT_OBJECT status 0x%x.\n",
			  status);
		   err = -EIO;
		}
             }
             kfree(mfs_import);
          }
          else {
             err = -ENOMEM;
          }
       }
       break;
    case MFS_IOCTL_LOW_ON_SPACE: /* crontab, ????? */
       fist_dprint(6, "mfs: MFS_IOCTL_LOW_ON_SPACE.\n");
       if (!suser()) {
          err = -EPERM;
       }
       else {
          mfs_status_t status;

	  status = mfs_do_request(stopd(file->f_dentry->d_inode->i_sb)
					->mfs_com_private,
				  MFS_COMMAND_SYSTEM_LOW_IN_SPACE,
				  NULL,
				  NULL,
				  NULL);
	  if (status != MFS_STATUS_OK) {
	     printk("mfs: LOW_IN_SPACE: mfsd answering a migrate with 0x%x.\n",
 		    status);
             err = -EIO;
	  }
       }
       break;

       /* mfs_migrate and mfs_unmigrate
        * BUGS
        *    - We do not do any access check. Either mfs_migrate/unmigrate
        *      should have some capabilities/be suid root and we should
        *      check cap/suser(), or we should do much testing here.
        *      This could include similar access right than for deletion,
        *      mfs migrated quota checks, etc.
        *    - We should really define more closely the semantics of this.
        */
    case MFS_IOCTL_MIGRATE: /* mfs_migrate */
       {
          mfs_status_t status;
          mfs_migrate_unmigrate_t *data;

          data = (mfs_migrate_unmigrate_t *)
                    kmalloc(sizeof(data->file_name), GFP_KERNEL);
          if (data) {
	     if (copy_from_user((void *) data,
			        (void *) arg,
			        sizeof(*data))) {
		err = -EFAULT;
	     }
             else {
		status = mfs_do_request(stopd(file->f_dentry->d_inode->i_sb)
					->mfs_com_private,
					MFS_COMMAND_MIGRATE_OBJECT,
					NULL,
					NULL,
					data);

		if (status != MFS_STATUS_OK) {
		   printk("mfs_ioctl: MFS_IOCTL_MIGRATE status 0x%x.\n",
			  status);
		   err = -EIO;
		}
             }

             kfree(data);
          }
          else {
             err = -ENOMEM;
          }
       }
       break;
    case MFS_IOCTL_UNMIGRATE: /* mfs_unmigrate */
       {
          mfs_status_t status;
          mfs_migrate_unmigrate_t *data;

          data = (mfs_migrate_unmigrate_t *)
                    kmalloc(sizeof(data->file_name), GFP_KERNEL);
          if (data) {
	     if (copy_from_user((void *) data,
			        (void *) arg,
			        sizeof(*data))) {
		err = -EFAULT;
	     }
             else {
		status = mfs_do_request(stopd(file->f_dentry->d_inode->i_sb)
					->mfs_com_private,
					MFS_COMMAND_UNMIGRATE_OBJECT,
					NULL,
					NULL,
					data);

		if (status != MFS_STATUS_OK) {
		   printk("mfs_ioctl: MFS_IOCTL_UNMIGRATE status 0x%x.\n",
			  status);
		   err = -EIO;
		}
             }

             kfree(data);
          }
          else {
             err = -ENOMEM;
          }
       }
       break;

    default:
	hidden_file = ftohf(file);
	/* pass operation to hidden filesystem, and return status */
	if (hidden_file->f_op && hidden_file->f_op->ioctl)
	    err = hidden_file->f_op->ioctl(itohi(inode), hidden_file, cmd, arg);
    } /* end of switch statement */

    print_exit_status(err);
    return err;
}


// STATIC int
// mfs_mmap(inode_t *inode, file_t *file, vm_area_t *vma)
STATIC int
mfs_mmap(file_t *file, vm_area_t *vma)
{
    int err = 0;
    file_t *hidden_file = ftohf(file);
    inode_t *inode = file->f_dentry->d_inode;
    inode_t *hidden_inode = itohi(inode);
    vm_area_t *hidden_vma = (vm_area_t *) 0xdeadc0de;

/* @@@ */
    ASSERT(hidden_file != NULL);
return hidden_file->f_op->mmap(hidden_file, vma);
/* @@@ */

    print_entry_location();

    fist_dprint(6, "MMAP1: inode 0x%x, hidden_inode 0x%x, inode->i_count %d, hidden_inode->i_count %d\n",
		(int) inode, (int) hidden_inode, (int) inode->i_count, (int) hidden_inode->i_count);

    if (!hidden_file->f_op || !hidden_file->f_op->mmap) {
	err = -ENODEV;
	goto out;
    }

    /*
     * Most of this code comes straight from generic_file_mmap
     * in mm/filemap.c.
     */
    if ((vma->vm_flags & VM_SHARED) && (vma->vm_flags & VM_MAYWRITE)) {
	vma->vm_ops = &mfs_shared_vmops;
	/* share_page() can only guarantee proper page sharing if * the offsets
	 * are all page aligned. */
	if (vma->vm_offset & (PAGE_SIZE - 1)) {
	    err = -EINVAL;
	    goto out;
	}
    } else {
	vma->vm_ops = &mfs_private_vmops;
	if (vma->vm_offset & (hidden_inode->i_sb->s_blocksize - 1)) {
	    err = -EINVAL;
	    goto out;
	}
    }
    if (!inode->i_sb || !S_ISREG(inode->i_mode)) {
	err = -EACCES;
	goto out;
    }
    if (!hidden_inode->i_op || !hidden_inode->i_op->readpage) {
	err = -ENOEXEC;
	goto out;
    }
    UPDATE_ATIME(inode);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,3)
    vma->vm_file = file;
    file->f_count++;
#endif

    /*
     * Now we do the hidden stuff, but only for shared maps.
     */
    if ((vma->vm_flags & VM_SHARED) && (vma->vm_flags & VM_MAYWRITE)) {
	hidden_vma = kmalloc(sizeof(vm_area_t), GFP_KERNEL);
	if (!hidden_vma) {
	    printk("MMAP: Out of memory\n");
	    err = -ENOMEM;
	    goto out;
	}
	// ION, is this right?
	memcpy(hidden_vma, vma, sizeof(vm_area_t));
	vmatohvma(vma) = hidden_vma;
	err = hidden_file->f_op->mmap(hidden_file, hidden_vma);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,3)
	hidden_vma->vm_file = hidden_file;
	hidden_file->f_count++;
#endif
    }

    /*
     * XXX: do we need to insert_vm_struct and merge_segments as is
     * done in do_mmap()?
     */
 out:
    fist_dprint(6, "MMAP2: inode 0x%x, hidden_inode 0x%x, inode->i_count %d, hidden_inode->i_count %d\n",
		(int) inode, (int) hidden_inode, (int) inode->i_count, (int) hidden_inode->i_count);
#if 0
    if (!err) {
	inode->i_mmap = vma;
	hidden_inode->i_mmap = hidden_vma;
    }
#endif
    print_exit_status(err);
    return err;
}


STATIC int
mfs_open(inode_t *inode, file_t *file)
{
    int err = -ENFILE;
    file_t *hidden_file;
    inode_t *hidden_inode = itohi(inode);
    dentry_t *hidden_dentry = mfs_hidden_dentry(file->f_dentry);

    print_entry_location();

    fist_print_dentry(__FUNCTION__ " IN hidden_dentry", hidden_dentry);

    hidden_file = get_empty_filp();
    if (!hidden_file)
	goto out;

    fist_dprint(8, "mfs_open: got empty hidden_file\n");
    file->private_data = hidden_file; /* link two file's */

    hidden_file->f_dentry = hidden_dentry;
    dget(hidden_dentry);

    if ((hidden_dentry->d_inode->i_size == 0)
        && (file->f_dentry->d_inode->i_size != 0)
        && !(MFS_IS_MIGRATED_FLAG(hidden_dentry->d_inode))) {
       /* See inode.c's mfs_lookup() for details */
       printk("mfs_open: BUG2.\n");
    }

retry:
    if (MFS_IS_MIGRATED_FLAG(hidden_dentry->d_inode)) {
       mfs_status_t status;

       fist_dprint(6, "mfs_open: %s is migrated.\n",
              file->f_dentry->d_name.name);

       /* TODO @@@
	*    - implement unmigrating or deletion depending on modes,
	*      properly.
        *    - either we do not migrate zero-sized files, or we
        *      optimize a little here.
        *    - see if other function (create(), truncate() need this too).
	*/
       if ((hidden_dentry->d_inode->i_size != 0)) {
          /* See inode.c's mfs_lookup() for details */
          printk("mfs_open: BUG.\n");
       }

       if (((file->f_mode & O_CREAT) && ! (file->f_mode & O_EXCL))
           || (file->f_mode & O_TRUNC)) {
          fist_dprint(6, "mfs_open: delete object.\n");
          /* File content will be suppressed */
	  status = mfs_do_request(stopd(file->f_dentry->d_inode->i_sb)
				  ->mfs_com_private,
				  MFS_COMMAND_DELETE_INDEX_OBJECT,
				  file->f_dentry, /* for the complete name */
				  NULL,
				  NULL);
          /* ignore the result */
       }
       else {
	  status = mfs_do_request(stopd(file->f_dentry->d_inode->i_sb)
				     ->mfs_com_private,
				  MFS_COMMAND_UNMIGRATE_OBJECT,
				  file->f_dentry, /* for the complete name */
				  NULL,
				  NULL);
	  if (status == MFS_STATUS_OK) {
	     MFS_CLEAR_MIGRATED_FLAG(hidden_dentry); /* @@@ what for */

	     if (MFS_IS_MIGRATED_FLAG(file->f_dentry->d_inode)) {
		/* Supposed to be done by mfsd */
		printk("mfs_open: BUG, still marked as migrated.\n");
	     }

	     if (MFS_IS_MIGRATED_FLAG(hidden_dentry->d_inode)) {
		/* Supposed to be done by mfsd */
		printk("mfs_open: BUG2, still marked as migrated.\n");
	     }

	     hidden_dentry->d_inode->i_size = file->f_dentry->d_inode->i_size;
	     fist_dprint(6, "mfs_open: %s is unmigrated.\n",
		    file->f_dentry->d_name.name);
	  }
	  else {
	     if (status == MFS_STATUS_OBJECT_NOT_FOUND) {
		printk("mfs_open: %s looks like migrated but isn't.\n",
		       file->f_dentry->d_name.name);
		/* TODO @@@
		 *    - reset migrated bit on the hidden_dentry->d_inode, and
		 *      make it go to disk so to clean things a bit.
		 *      could use a macro since probably needed for unmigrate
		 *      too. Or could use a function since shared with inode.c
		 *    - also reset on dentry->d_inode migrated flag.
		 *    - ensure that all people in the queue get awaken before
		 *      we start to say not found. If this not done, this
		 *      can happen.
		 * NOTES
		 *    - Will happen for all queued requests while unmigrating.
		 *      Maybe check again the bits and only display the
		 *      warnings if still looks migrated.
		 */
		/* Let normal open proceed, even if this is buggy */
	     }
	     else {
		printk("mfs_open: mfsd error 0x%x.\n", status);
		/* In that case, we leave the migrated bit and we
		 * keep the zero size. open() is free to fail, and
		 * stat() will return zero size.
		 * @@@
		 */
		err = -EIO;
		goto out;
	     }
	  }
       }
    }

    if (MFS_IS_MIGRATED_FLAG(hidden_dentry->d_inode)) {
       /* Could happen if object was migrated again just before we were
        * awakened.
        */
       printk("mfs_open: migrated, trying again.\n");
       goto retry;
    }

    fist_dprint(6, "mfs_open: got f_inode\n");
    hidden_file->f_mode = file->f_mode | FMODE_READ;
    hidden_file->f_flags = file->f_flags;
    if (file->f_flags & O_APPEND) {
	fist_dprint(5, "file is opened in append-only mode!!!\n");
	hidden_file->f_flags &= ~O_APPEND;	/* turn off O_APPEND flag */
    }
    hidden_file->f_pos = file->f_pos;
    hidden_file->f_reada = file->f_reada;

    hidden_file->f_op = NULL;
    if (hidden_inode->i_op)
	hidden_file->f_op = hidden_inode->i_op->default_file_ops;

    fist_dprint(6, "mfs_open: pre open\n");
    if (hidden_file->f_op && hidden_file->f_op->open) {
	err = hidden_file->f_op->open(hidden_inode, hidden_file);
	if (err) {
	    printk(KERN_WARNING "mfs_open: error(%d)\n", err);
	    hidden_file->f_dentry = NULL;
	    put_filp(hidden_file);
	    goto out;
	}
#if 0
        else {
           printk("mfs_open: %d / %d\n",
                  inode->i_writecount,
                  hidden_inode->i_writecount);
	}
#endif
    } else {
	err = 0;
    }

    hidden_file->f_flags &= ~(O_CREAT | O_EXCL | O_NOCTTY | O_TRUNC);

    fist_checkinode(inode, "post mfs_open");
    fist_dprint(6, "mfs_open: done (%d)\n", err);

out:
    fist_print_dentry(__FUNCTION__ " OUT hidden_dentry", hidden_dentry);
    print_exit_status(err);
    return err;
}


// newly implemented
STATIC int
mfs_flush(file_t *file)
{
    int err = 0;		/* assume ok (see open.c:close_fp) */
    file_t *hidden_file = ftohf(file);

    print_entry_location();

    if (!hidden_file->f_op || !hidden_file->f_op->flush)
	goto out;

    err = hidden_file->f_op->flush(hidden_file);

out:
    print_exit_status(err);
    return err;
}


// STATIC void
STATIC int
mfs_release(inode_t *inode, file_t *file)
{
    int err = 0;
    file_t *hidden_file = ftohf(file);
    dentry_t *hidden_dentry;

    print_entry_location();
    ASSERT(hidden_file != NULL);

    mfs_notify_dies(file->f_dentry,
                    stopd(file->f_dentry->d_inode->i_sb)->mfs_com_private);

    fist_print_dentry("mfs_release IN hidden_dentry", hidden_file->f_dentry);
    fist_checkinode(inode, "mfs_release");
    fist_dprint(6, "mfs_release IN, file->f_count=%d\n", file->f_count);
    /*
     * will decrement file refcount, and if 0, destroy the file,
     * which will call the lower file system's file release function.
     */

    hidden_dentry = hidden_file->f_dentry;

#if 0 /* doesn't seem to help, infact */
    if (hidden_file->f_op && hidden_file->f_op->release) {
       hidden_file->f_op->release(hidden_dentry->d_inode,
                                  hidden_file);
    }
#endif

    fput(hidden_file);

    fist_dprint(6, "mfs_release done\n");
    fist_checkinode(inode, "post mfs_release");

    fist_print_dentry("mfs_release OUT hidden_dentry", hidden_dentry);
    print_exit_status(err);
    return err;
}


// STATIC int
// mfs_fsync(inode_t *inode, file_t *file)
STATIC int
mfs_fsync(file_t *file, dentry_t *dentry)
{
    int err = -EINVAL;
    file_t *hidden_file = ftohf(file);
    dentry_t *hidden_dentry = mfs_hidden_dentry(dentry);

    print_entry_location();

    if (hidden_file->f_op && hidden_file->f_op->fsync) {
	down(&hidden_dentry->d_inode->i_sem);
	err = hidden_file->f_op->fsync(hidden_file, hidden_dentry);
	up(&hidden_dentry->d_inode->i_sem);
    }

    print_exit_status(err);
    return err;
}


// STATIC int
// mfs_fasync(inode_t *inode, file_t *file, int flag)
STATIC int
mfs_fasync(int fd, file_t *file, int flag)
{
    int err = 0;
    file_t *hidden_file = ftohf(file);

    print_entry_location();

    if (hidden_file->f_op && hidden_file->f_op->fasync)
	err = hidden_file->f_op->fasync(fd, hidden_file, flag);

    print_exit_status(err);
    return err;
}


// newly implemented
STATIC int
mfs_check_media_change(kdev_t dev)
{
    int err = 0;
    super_block_t *sb;

    print_entry_location();

    sb = get_super(dev);
    err = check_disk_change(stohs(sb)->s_dev);

    print_exit_status(err);
    return err;
}


#ifdef NOT_NEEDED
/*
 * This is not needed, because our check_media_change will call
 * the generic check_disk_change() which will run the revalidate
 * function of the hidden file system, if implemented.
 */
STATIC int
mfs_file_revalidate(kdev_t dev)
{
    return -ENOSYS;
}
#endif /* NOT_NEEDED */

// newly implemented
STATIC int
mfs_lock(file_t *file, int cmd, struct file_lock *fl)
{
    int err = 0;
    file_t *hidden_file = ftohf(file);

    print_entry_location();

    if (hidden_file->f_op->lock) {
	err = hidden_file->f_op->lock(hidden_file, F_GETLK, fl);
    } else {
	posix_test_lock(hidden_file, fl);
    }

    print_exit_status(err);
    return err;
}




struct file_operations mfs_fops =
{
    mfs_llseek,		/* llseek */
#if 1
    generic_file_read,		/* read */
#else
    mfs_read,		/* read */
#endif
#if 0
    generic_file_write,		/* write */
#else
    mfs_write,		/* write */
#endif
    mfs_readdir,		/* readdir */
    mfs_poll,		/* poll */
    mfs_ioctl,		/* ioctl */
    mfs_mmap,		/* mmap */
    mfs_open,		/* open */
    mfs_flush,		/* flush */
    mfs_release,		/* release */
    mfs_fsync,		/* fsync */
    mfs_fasync,		/* fasync */
    mfs_check_media_change,	/* check_media_change */
#ifdef NOT_NEEDED
    mfs_file_revalidate,	/* revalidate */
#else /* not NOT_NEEDED */
    NULL,			/* revalidate */
#endif /* not NOT_NEEDED */
    mfs_lock			/* lock */
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */




