/*
 *  $Id: mfs.h,v 1.3 1999/10/19 09:45:46 schaefer Exp $
 */

#ifndef __FIST_MFS_H_
#define __FIST_MFS_H_

/*
 * STRUCTURES:
 */

/* mfs inode data in memory */
struct mfs_inode_info {
    inode_t *inode;
};

/* mfs super-block data in memory */
struct mfs_sb_info {
    super_block_t *s_sb;
#if 0
    /* not needed b/c linux has an s_root field in struct super_block */
    dentry_t *s_root;
#endif
    // inode_t *s_rootinode;

   void *mfs_com_private; /* for each filesystem */
};

/*
 * MACROS:
 */
// File TO Hidden File
#define ftohf(file) ((file_t *)((file)->private_data))
// Inode TO Private Data
#define itopd(ino) ((struct mfs_inode_info *)(ino)->u.generic_ip)
// Inode TO Hidden Inode
#define itohi(ino) (itopd(ino)->inode)
// Superblock TO Private Data
#define stopd(super) ((struct mfs_sb_info *)(super)->u.generic_sbp)
// Superblock TO Hidden Superblock
#define stohs(super) (stopd(super)->s_sb)
// Dentry TO Hidden Dentry (and private data --- for the future)
#define dtopd(dentry) ((dentry)->d_fsdata)
#define dtohd(dentry) ((dentry_t *)dtopd(dentry))
/* VMArea TO Hidden VMArea */
#define vmatohvma(vma) ((vm_area_t *)((vma)->vm_private_data))

#define sbt(sb) ((sb)->s_type->name)

/*
 * EXTERNALS:
 */
extern struct file_operations mfs_fops;
extern struct inode_operations mfs_iops;
extern struct super_operations mfs_sops;
extern struct dentry_operations mfs_dops;
extern struct vm_operations_struct mfs_shared_vmops;
extern struct vm_operations_struct mfs_private_vmops;

extern int mfs_decode_block(const char *inbuf, char *outbuf, int len);
extern int mfs_encode_block(const char *inbuf, char *outbuf, int len);
extern int mfs_interpose(dentry_t *hidden_dentry, dentry_t *this_dentry, super_block_t *sb, int flag);

#ifdef FIST_DEBUG
# define mfs_hidden_dentry(d) __mfs_hidden_dentry(__FILE__,__FUNCTION__,__LINE__,(d))
extern dentry_t *__mfs_hidden_dentry(char *file, char *func, int line, dentry_t *this_dentry);
#else /* not FIST_DEBUG */
# define mfs_hidden_dentry(dentry) dtohd(dentry)
#endif /* not FIST_DEBUG */

#endif	/* __FIST_MFS_H_ */

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
