/* Project mfs --- Migration file system
 * mfs_user_ioctls.h -- User/kernel communication definitions
 * Creation: 11/12/1999
 * Update:   11/12/1999
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This include file defines data types used in the mfsd/mfs communication.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    11/12/99  schaefer  Created this file.
 * $Id: mfs_user_ioctls.h,v 1.8 2000/02/24 08:14:10 schaefer Exp $
 */

#ifndef _MFS_USER_IOCTLS_H
#define _MFS_USER_IOCTLS_H

/* IOCTL definitions */

/* NOTES
 *    - It is important, when using a structure as parameter, that
 *      all the types used in the structure are the same in kernel
 *      driver and in user-space utilities. That's why we
 *      define our own types.
 *    - See also mfs_user_ioctls.h
 * BUGS
 *    - We should allocate those so that no other fs ioctl ressembles ours.
 *    - Must cheat with _IOR since opening read only "/".
 */
#define MFS_IOCTL_IMPORT_OBJECT _IOR(0x16, 1, mfs_import_t) /* root only */
#define MFS_IOCTL_LOW_ON_SPACE  _IOR(0x16, 2, NULL) /* root only */

#define MFS_IOCTL_MIGRATE       _IOR(0x16, 3, mfs_migrate_unmigrate_t) /* everyone */
#define MFS_IOCTL_UNMIGRATE     _IOR(0x16, 4, mfs_migrate_unmigrate_t) /* everyone */

#define MFS_IOCTL_DEBUG         _IOR(0x16, 5, int)

/* Type definitions */

typedef struct {
   char file_name[MAX_PATHNAME_LENGTH]; /* relative to mfs */
} mfs_migrate_unmigrate_t;

typedef struct {
   mfs_import_object_locator_t locator; /* private to unmigration method */

   mfs_index_object_size_t size;
   mfs_uid_t uid;
   mfs_gid_t gid;
   mfs_time_t mtime; /* @@@ others? @@@ */
   mfs_mode_t mode;

   char file_name[MAX_PATHNAME_LENGTH];

   char migrated_on[MAX_MIGRATED_ON_PATHNAME_LENGTH];
} mfs_import_t;

#endif /* !_MFS_USER_IOCTLS_H */




