/* Project mfs --- Migration file system
 * changer_interface.h -- Generic changer device
 * Creation: 02/03/2000
 * Update:   02/03/2000
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This include file defines the data types and functions used for
 *    communication with the generic changer device. The idea being
 *    that requests are queued to a single generic changer device
 *    (or one per changer hardware, e.g.), which loads the appropriate
 *    object (tape, MO, etc) into one of the drives (possibly unmounting
 *    first), and then fork()s a tape reader or writer process which will
 *    be connected to the program using the changer via fds.
 * NOTES
 *    - Still need to define a bit more the communication methods.
 *      pipes, sockets, etc.
 *    - Instead of giving the data through the data channel, the tape
 *      process could untar all the files itself. However, we prefer
 *      to do it ourselves (e.g. so that we can react correctly on
 *      a no space condition). This means that like tapefs, the tape
 *      process needs to understand bits of the tar structure.
 *    - If the drive is a MO drive, you simply get a fd on the
 *      file itself.
 * BUGS
 * TODO
 *    - with inclusion of mfs_definitions.h, rework this file.
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    02/03/00  schaefer  Created this file.
 * $Id: changer_interface.h,v 1.2 2000/03/04 13:28:12 schaefer Exp $
 */

#ifndef _CHANGER_INTERFACE_H
#define _CHANGER_INTERFACE_H

/* Type definitions */

#define CHANGER_REQUEST_PATHLEN 1024
typedef unsigned long int changer_user_data_t;
typedef unsigned long int changer_locator_t;

typedef enum {
   CHANGER_OPERATION_RETRIEVE,
   CHANGER_OPERATION_WRITE
} changer_operation_t;

typedef struct {
   changer_user_data_t user_private;

   char container_name[CHANGER_REQUEST_PATHLEN]; /* E.g.: tape label */
   changer_locator_t locator; /* For example: file number and record number
                               * on tape.
                               */
   char object_name[CHANGER_REQUEST_PATHLEN]; /* Normally the filepath/name */
   changer_operation_t operation;
} changer_request_t;

typedef enum {
   CHANGER_RESULT_OK_DATA_COMING,
   CHANGER_RESULT_OK_DATA_GOING,
   CHANGER_RESULT_ERROR_CANT_LOCATE_CONTAINER,
   CHANGER_RESULT_ERROR_CANT_LOCATE_OBJECT,
   CHANGER_RESULT_ERROR_GENERIC /* see errno */
} changer_result_t;

typedef unsigned long int changer_result_qualifier_t;

typedef struct {
   changer_user_data_t user_private; /* from changer_request_t, allows to
                                      * match requests.
                                      */
   changer_result_t result;
   changer_result_qualifier_t qualifier;
   /* the data is coming afterwards, or is to be sent then ? */
} changer_answer_t;

#endif /* !_CHANGER_INTERFACE_H */

