/* Project mfs --- Migration file system
 * index_interface.h -- Index management routine definitions.
 * Creation: 19/12/1999
 * Update:   19/12/1999
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This include file defines data types and functions used for
 *    index management. There is only one index management per
 *    link (compile object). This interfaces with the dcache
 *    module.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    19/12/99  schaefer  Created this file.
 * $Id: index_interface.h,v 1.4 1999/12/30 10:04:16 schaefer Exp $
 */

#ifndef _INDEX_INTERFACE_H
#define _INDEX_INTERFACE_H

/* Type definitions */

/* Function definitions */

/* NAME
 *    index_init
 * DESCRIPTION
 *    Initializes the index, possibly creating or reading from disk
 *    the index information from the passed file name.
 * RESULT
 *    TRUE if ok
 * NOTES
 * BUGS
 * TODO
 */
BOOL index_init(const char *name);

/* NAME
 *    index_deinit
 * DESCRIPTION
 *    Deinitializes the index, possibly saving to disk
 *    the index information from the associated name.
 * RESULT
 *    TRUE if ok
 * NOTES
 * BUGS
 * TODO
 */
BOOL index_deinit(void);

/* NAME
 *    index_add
 * DESCRIPTION
 *    Adds an entry to the index.
 * RESULT
 *    TRUE if ok
 * NOTES
 * BUGS
 *    - Error code should be enhanced.
 * TODO
 */
BOOL index_add(const char *name, mfs_import_data_t *data);

/* NAME
 *    index_remove
 * DESCRIPTION
 *    Removes an entry from the index.
 * RESULT
 *    TRUE if ok
 * NOTES
 * BUGS
 * TODO
 */
BOOL index_remove(const char *name);

/* NAME
 *    index_get
 * DESCRIPTION
 *    Getss an entry to the index.
 * RESULT
 *    TRUE if ok
 * NOTES
 * BUGS
 *    - Error code should be enhanced.
 *    - Not symetric (in arguments)
 * TODO
 */
BOOL index_get(const char *name, mfs_import_data_t *answer);

#endif /* !_INDEX_INTERFACE_H */

