/* Project mfs --- Migration file system
 * lib_interface.h -- Utility library function definitions.
 * Creation: 04/01/2000
 * Update:   04/01/2000
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This include file defines data types and functions exported by
 *    the utility library.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    04/01/00  schaefer  Created this file.
 * $Id: lib_interface.h,v 1.3 2000/01/10 17:40:23 schaefer Exp $
 */

#ifndef _LIB_INTERFACE_H
#define _LIB_INTERFACE_H

/* Type definitions */

#define EXIT_ERROR_INTERNAL_ERROR 4
#define EXIT_ERROR_INVALID_PARAMETERS 3
#define EXIT_ERROR_BAD_MOUNT_POINT 2
#define EXIT_ERROR_NOT_ALL_OK 1
#define EXIT_ERROR_NONE 0

/* Function definitions */

/* NAME
 *    mfsd_open_handle
 * DESCRIPTION
 *    Opens a file descriptor on the mfsd handle for further communication.
 * RESULT
 *    file descriptor (fd_t) or -1 if cannot be opened
 * NOTES
 *    - Able to open it doesn't mean there is a mfsd connected.
 *    - Currently implemented by opening the directory which the file_name
 *      given is in.
 *    - You can close the returned file handle normally with close().
 * BUGS
 * TODO
 */
fd_t mfsd_open_handle(char *file_name);

/* NAME
 *    mfs_migrate_unmigrate
 * DESCRIPTION
 *    Migrates or unmigrates the specified file.
 * RESULT
 *    TRUE if succeeded (name fits and operation succeeds).
 * NOTES
 * BUGS
 * TODO
 *    - Add more details on asynchronicity, error codes, etc.
 */
BOOL mfs_migrate_unmigrate(fd_t fd,
                           const char *file,
                           BOOL migrate,
                           const char *current_directory_path);
#endif /* !_LIB_INTERFACE_H */

