/* Project mfs --- Migration file system
 * main.c -- Utility library functions
 * Creation: 04/01/2000
 * Update:   04/01/2000
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This implements the general functions for the utility library.
 * NOTES
 * BUGS
 * TODO
 *    - Each function should be in separate .o file for not wasting
 *      space.
 *    - Maybe implement a shared library ?
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    04/01/00  schaefer  Created this file.
 * $Id: main.c,v 1.5 2000/01/10 17:40:24 schaefer Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "../../mfs/mfs_definitions.h"
#include "../../mfs/mfs_user_ioctls.h"
#include "lib_interface.h"

/* Function implementation */

fd_t mfsd_open_handle(char *file_name) {
   char buffer[MAX_PATHNAME_LENGTH];
   char *last = strrchr(file_name, '/');

   if (last) {
      if ((last - file_name) < sizeof(buffer)) {
         memcpy(buffer, file_name, (last - file_name));
         buffer[last - file_name] = '\0';
      }
      else {
         fprintf(stderr, "mfsd_open_handle: buffer overflow (%s).\n",
                 file_name);
         return (-1);
      }
   }
   else {
      memcpy(buffer, ".", 1);
   }

   return open(buffer, O_RDONLY);
}

BOOL mfs_migrate_unmigrate(fd_t fd,
                           const char *file,
                           BOOL migrate,
                           const char *current_directory_path) {
   mfs_migrate_unmigrate_t temp;

   if (snprintf(temp.file_name,
                sizeof(temp.file_name),
                "%s%s%s",
                (file[0] == '/')
                   ? "\0"
                   : current_directory_path,
                (file[0] == '/')
                   ? "\0"
                   : "/",
                file)) {
      if (ioctl(fd,
                migrate
                   ? MFS_IOCTL_MIGRATE
                   : MFS_IOCTL_UNMIGRATE,
                &temp)) {
	 perror("mfs_migrate_unmigrate: ioctl()");
	 return FALSE;
      }
      else {
	 return TRUE;
      }
      /* NOT REACHED */
   }
   else {
      fprintf(stderr, "mfs_migrate_unmigrate: pathname too long %s %s.\n",
                      file,
                      current_directory_path);
      return FALSE;
   }
   /* NOT REACHED */
}

