/* Project mfs --- Migration file system
 * main.c -- mfs_migrate main module
 * Creation: 04/01/2000
 * Update:   04/01/2000
 * V1.0 PV001
 * AUTHOR
 *    Marc SCHAEFER <schaefer@alphanet.ch>
 * LICENSE
 *    GPL
 * DESCRIPTION
 *    This implements the command line parsing and initialization for the
 *    mfs_migrate tool. The mfs_migrate tool allows to migrate a file
 *    to offline storage.
 *    The mfs filesystem must be mounted, and ioctl()s will be used
 *    to communicate the information to the mfsd process through the mfs
 *    filesystem.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 *    04/01/00  schaefer  Created this file.
 * $Id: main.c,v 1.4 2000/01/10 17:40:24 schaefer Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "../../mfs/mfs_definitions.h"
#include "../../mfs/mfs_user_ioctls.h"
#include "../lib/lib_interface.h"

/* Private function definitions */

/* NAME
 *    usage
 * DESCRIPTION
 *    Prints a short usage summary on stderr.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void usage(char *program_name);

/* Function implementation */

int main(int argc, char **argv) {
   int result = EXIT_ERROR_INVALID_PARAMETERS;

   if (argc >= 2) {
      char current_directory[MAX_PATHNAME_LENGTH];
      fd_t f = mfsd_open_handle(argv[1]);
   
      if (f != -1) {
         if (getcwd(current_directory, sizeof(current_directory))) {
            int i = 1;

	    result = EXIT_ERROR_NONE;

	    while (i < argc) {
	       int temp_result;

	       temp_result = mfs_migrate_unmigrate(f,
						   argv[i],
						   TRUE,
						   current_directory)
				? EXIT_ERROR_NONE
				: EXIT_ERROR_NOT_ALL_OK;

	       i++;
	       if (result < temp_result) {
		  result = temp_result;
	       }
	    }
         }
         else {
            perror("getcwd()");
            result = EXIT_ERROR_INTERNAL_ERROR;
         }
         close(f);
      }
      else {
         perror("open()");
         result = EXIT_ERROR_BAD_MOUNT_POINT;
      }
   }

   if (result == EXIT_ERROR_INVALID_PARAMETERS) {
      usage(argv[0]);
   }

   exit(result);
}

static void usage(char *program_name) {
   fprintf(stderr,
           "%s file ... file\n",
           program_name);
   fprintf(stderr, "%s: bad args.\n", program_name);
}
